import asyncio
import logging

from pyee import AsyncIOEventEmitter
from fastapi.encoders import jsonable_encoder

from .registry import R


def responder(event):
    def wrapper(original):
        R.add(event, original)
        return original

    return wrapper


class NamedEventEmitter(AsyncIOEventEmitter):
    def __init__(self, name, *args, **kwargs):
        self.name = name

        AsyncIOEventEmitter.__init__(self, *args, **kwargs)


events = NamedEventEmitter(name="events")
actions = NamedEventEmitter(name="actions")
commands = NamedEventEmitter(name="commands")


def emit(emitter: NamedEventEmitter, event, payload):
    async def _emit_async():
        emitter.emit(event, jsonable_payload)

    jsonable_payload = jsonable_encoder(payload)
    log = logging.getLogger(__name__)
    log.info(f"Emitting '{event}' using emitter '{emitter.name}'")
    log.debug(jsonable_payload)

    loop = asyncio.get_event_loop()
    loop.create_task(_emit_async())



import typing

from starlette.responses import Response


class R:
    callbacks = dict()

    @classmethod
    def add(cls, event, handler: typing.Callable[[dict], Response]):
        cls.callbacks[event] = handler

    @classmethod
    def handle(cls, event: str, payload: dict):
        return cls.callbacks[event](payload)



import json
import logging
import typing

from fastapi import APIRouter, Depends
from starlette.requests import Request
from starlette.responses import Response
from starlette.status import HTTP_200_OK

from slackers.hooks import actions, commands, emit, events
from slackers.models import SlackAction, SlackChallenge, SlackCommand, SlackEnvelope
from slackers.registry import R
from slackers.verification import check_timeout, verify_signature

log = logging.getLogger(__name__)

router = APIRouter()


@router.post(
    "/events",
    status_code=HTTP_200_OK,
    dependencies=[Depends(verify_signature), Depends(check_timeout)],
)
async def post_events(message: typing.Union[SlackEnvelope, SlackChallenge]):
    if isinstance(message, SlackChallenge):
        return message.challenge

    emit(events, message.event["type"], payload=message)
    return Response()


@router.post(
    "/actions",
    status_code=HTTP_200_OK,
    dependencies=[Depends(verify_signature), Depends(check_timeout)],
)
async def post_actions(request: Request) -> Response:
    form = await request.form()
    form_data = json.loads(form["payload"])

    # have the convenience of pydantic validation
    action = SlackAction(**form_data)
    _events = [action.type]
    if action.actions:
        _events.extend(_add_action_triggers(action))
    if action.callback_id:
        _events.append(f"{action.type}:{action.callback_id}")
    if action.view:
        view_callback_id = action.view.get("callback_id")
        if view_callback_id:
            _events.append(f"{action.type}:{view_callback_id}")

    for _event in _events:
        emit(actions, _event, payload=action)

    registered_handlers = set(R.callbacks.keys())
    registered_events = set(_events)
    handlers_to_call = registered_handlers.intersection(registered_events)
    if len(handlers_to_call) > 1:
        raise ValueError(f"Multiple response handlers found.")

    if handlers_to_call:
        handle = handlers_to_call.pop()
        response = R.handle(handle, action.dict())
        assert isinstance(
            response, Response
        ), "Please return a starlette.responses.Response"
        return response
    return Response()


def _add_action_triggers(action: SlackAction) -> list:
    gathered_events = [
        f"{action.type}:{triggered_action['action_id']}"
        for triggered_action in action.actions
        if "action_id" in triggered_action
    ]
    gathered_events.extend(
        [
            f"{action.type}:{triggered_action['name']}"
            for triggered_action in action.actions
            if "name" in triggered_action
        ]
    )
    gathered_events.extend(
        [
            f"{action.type}:{triggered_action['type']}"
            for triggered_action in action.actions
            if "type" in triggered_action
        ]
    )
    gathered_events.extend(
        [
            f"{action.type}:{triggered_action['name']}:{triggered_action['type']}"
            for triggered_action in action.actions
            if "name" in triggered_action and "type" in triggered_action
        ]
    )

    return gathered_events


@router.post(
    "/commands",
    status_code=HTTP_200_OK,
    dependencies=[Depends(verify_signature), Depends(check_timeout)],
)
async def post_commands(request: Request):
    form = await request.form()
    command = SlackCommand(**form)
    emit(commands, command.command.lstrip("/"), command)

    return Response()



import hmac
import math
import os
import time
import logging

from hashlib import sha256

from fastapi import Header
from starlette.status import HTTP_403_FORBIDDEN
from starlette.requests import Request
from starlette.exceptions import HTTPException


log = logging.getLogger(__name__)


async def verify_signature(
    request: Request,
    x_slack_signature: str = Header(...),
    x_slack_request_timestamp: str = Header(...),
):

    log.debug("Starting verification")

    body = await request.body()
    to_verify = str.encode("v0:" + str(x_slack_request_timestamp) + ":") + body
    our_hash = hmac.new(
        os.environ.get("SLACK_SIGNING_SECRET").encode(), to_verify, sha256
    ).hexdigest()
    our_signature = "v0=" + our_hash

    if not hmac.compare_digest(x_slack_signature, our_signature):
        log.info("Slack verification failed")
        raise HTTPException(HTTP_403_FORBIDDEN, "Forbidden")

    log.debug("Verification successful")


def check_timeout(x_slack_request_timestamp: str = Header(...)):
    timeout = 60 * 5  # 5 minutes
    request_timeout_time = int(x_slack_request_timestamp) + timeout
    current_time = math.ceil(time.time())

    if current_time > request_timeout_time:
        log.info("Slack request timestamp reached timeout")
        raise HTTPException(HTTP_403_FORBIDDEN, "Forbidden")



from typing import List, Optional

from pydantic import BaseModel


class SlackBase(BaseModel):
    token: str


class SlackChallenge(SlackBase):
    challenge: str
    type: str


class SlackEnvelope(SlackBase):
    class Config:
        extra = "allow"

    team_id: str
    api_app_id: str
    event: dict
    type: str
    event_id: str
    event_time: int
    authorizations: Optional[List[dict]]
    is_ext_shared_channel: Optional[bool]
    event_context: Optional[str]


class SlackAction(SlackBase):
    class Config:
        extra = "allow"

    type: str

    actions: list = None
    api_app_id: str = None
    callback_id: str = None
    channel: dict = None
    container: dict = None
    hash: str = None
    is_cleared: bool = None
    message: dict = None
    response_url: str = None
    team: dict = None
    trigger_id: str = None
    user: dict = None
    view: dict = None


class SlackCommand(SlackBase):
    command: str
    response_url: str
    trigger_id: str
    user_id: str
    user_name: str
    team_id: str
    channel_id: str
    text: str
